/* ******************************************************************************

	FILE:			SSBaseApp.cpp
	AUTHOR:			Jean-Sbastien Perrier

	DESCRIPTION:	Base class for an OpenGL screen saver application. This
					class handles everything a screen saver should do. It creates
					the window and handles the OpenGL context creation. A lot of 
					things have been hard coded in order to simplify the class 
					(such as the pixel format and frame buffer configuration for
					the Opengl context).

	$History::                                                                  $

****************************************************************************** */

// Includes ****************************************************************** //
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <regstr.h>
#include <stdlib.h>		// for atoi
#include <string.h>		// for strcpy
#include <math.h>		// for abs

#include "SSBaseApp.h"
#include "SSUtils.h"

// Defines ******************************************************************* //
// Standard screen saver path for common settings.
#define REGSTR_PATH_PLUSSCR (REGSTR_PATH_SETUP "\\Screen Savers")

// Types ********************************************************************* //

// Consts ******************************************************************** //

// Globals ******************************************************************* //

// Static Members Storage Point ********************************************** //
// Global instance of the application (Like a singleton).
SSBaseApp *SSBaseApp::s_pInstance = NULL;

// Function Prototypes ******************************************************* //

// Module External Functions ************************************************* //


/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:	SSBaseApp
// DESCRIPTION:	Class constructor.
// PARAMETERS:	(in) a_hInstance	-	Handle to the application's instance.
// REMARKS:		None
/////////////////////////////////////////////////////////////////////////////////
SSBaseApp::SSBaseApp(HINSTANCE a_hInstance) :
	m_hInstance(a_hInstance),
	// Setup default values.
	m_hWnd(NULL),	
	m_bReallyClose(FALSE),
	m_bIsDialogActive(FALSE),
	m_dwPasswordDelay(15),
	m_dwMouseThreshold(50),
	m_bMuteSound(0),
	m_eMode(smNone)
{
	// Set the current global instance
	SSBaseApp::SetGlobalInstance(this);

	// Setup the default names. These can be overriden by
	// the derived classes.
	strcpy(m_szWindowName, "ScreenSaverWnd");
	strcpy(m_szWindowClassName, "ScreenSaverWndClass");
	
	// Read the General screen saver settings from the registry.
	SSUtlReg Reg;
	if(Reg.Open(REGSTR_PATH_PLUSSCR))
	{
		m_dwPasswordDelay = Reg.ReadDWord("Password Delay", m_dwPasswordDelay);
		m_dwMouseThreshold = Reg.ReadDWord("Mouse Threshold", m_dwMouseThreshold);
		m_bMuteSound = Reg.ReadDWord("Mute Sound", m_bMuteSound);
		Reg.Close();
	}
}

/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:	ProcessCommandLine
// DESCRIPTION:	Parse the command line and setup the application's mode and
//				get the window handle, if any.
// PARAMETERS:	None
// RETURN VALUE: Window handle.
// REMARKS:		This function will set the application's mode.
/////////////////////////////////////////////////////////////////////////////////
HWND SSBaseApp::ProcessCommandLine(char *a_szCmd)
{
	HWND hParentWnd=NULL;		// Parent window's handle.
	char szArg[256];			// Buffer to hold the isolated argument
	char *szLimit = " ";		// Default command line delimiter
	char *szCmd = a_szCmd;		// Pointer to command line string.

	// We read the first argument, this
	// should be the name of the program.
	szCmd = SSUtl_GetArg(szCmd, szArg, szLimit);

	if(!szCmd)
	{	// No more argument, then we are in configuration mode.
		m_eMode = smConfig;
	}
	else
	{	// Get the second argument
		szCmd = SSUtl_GetArg(szCmd, szArg, szLimit);
		switch(szArg[1])
		{
		// PREVIEW MODE
		case 'p':	case 'P':
		case 'l':	case 'L':
			m_eMode = smPreview;
			// Get the window's handle
			szCmd = SSUtl_GetArg(szCmd, szArg, szLimit);
			hParentWnd = (HWND)atoi(szArg);
			break;
		// SCREEN SAVER MODE
		case 's':	case 'S':	
			m_eMode = smSaver;	
			break;
		// CONFIGURATION MODE
		case 'c':	case 'C':
			m_eMode = smConfig;
			// Get window's handle
			szCmd = SSUtl_GetArg(szCmd, szArg, " :");
			if(*szArg)	hParentWnd = (HWND)atoi(szArg);
			else		hParentWnd = GetForegroundWindow(); 
			break;
		// PASSWORD MODE
		case 'a':	case 'A':
			szCmd = SSUtl_GetArg(szCmd, szArg, " :");
			hParentWnd = (HWND)atoi(szArg);
			m_eMode = smPassword;
			break;
		}
	}
	return hParentWnd;
}

void SSBaseApp::CloseSaverWindow()
{
	m_bReallyClose = true;
	PostMessage(m_hWnd,WM_CLOSE,0,0);
}

void SSBaseApp::StartDialog()
{
	m_bIsDialogActive = true;
	SendMessage(m_hWnd,WM_SETCURSOR,0,0);
}

void SSBaseApp::EndDialog()
{
	m_bIsDialogActive = false;
	SendMessage(m_hWnd,WM_SETCURSOR,0,0);
	GetCursorPos(&m_InitCursorPos);
}

void SSBaseApp::OnCreate(HWND a_hWnd)
{
	SetWindowHandle(a_hWnd);
	GetCursorPos(&m_InitCursorPos);
	m_dwInitTime = GetTickCount();
}


/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:	RunScreenSaver
// DESCRIPTION:	This functions is very central to the whole screen saver
//				application. It creates the windows and opengl context,
//				then run the main loop and finally, clean up and return.
// PARAMETERS:	(in) a_hParentWnd	-	Handle the parent window (can be NULL).
//				(in) nCmdShow	-	Sent by windows on creation.
// RETURN VALUE:None
// REMARKS:		None
/////////////////////////////////////////////////////////////////////////////////
void SSBaseApp::RunScreenSaver(HWND a_hParentWnd, int nCmdShow)
{
	//-----------------------------------------------------
	// STEP 1 - REGISTER WINDOW CLASS
	//-----------------------------------------------------
	WNDCLASS wc;
	wc.style=CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc	= SaverWindowProc;
	wc.cbClsExtra	= 0;
	wc.cbWndExtra	= 0;
	wc.hInstance	= m_hInstance;
	wc.hIcon		= NULL;
	wc.hCursor		= NULL;
	wc.hbrBackground = NULL;
	wc.lpszMenuName	= NULL;
	wc.lpszClassName	= GetWindowClassName();

	if(!RegisterClass(&wc)) 
	{
		MessageBox(NULL, "RegisterClass() failed:  "
			"Cannot register window class.", "Error", MB_OK);
		return;
	}

	//-----------------------------------------------------
	// STEP 2 - CREATE MAIN WINDOW
	//-----------------------------------------------------
	int cx, cy;
	HWND hParentWnd = NULL;
	DWORD dwStyle;

	if(GetMode() == smPreview)
	{	
		RECT rc;
		GetWindowRect(a_hParentWnd,&rc);
		cx = rc.right - rc.left;
		cy = rc.bottom - rc.top;
		dwStyle = WS_CHILD|WS_VISIBLE;
		hParentWnd = a_hParentWnd;

	}
	else
	{
		cx = GetSystemMetrics(SM_CXSCREEN);
		cy = GetSystemMetrics(SM_CYSCREEN);		
		dwStyle = WS_POPUP|WS_VISIBLE;
	}

	// Create the window for the screen saver
	HWND hScrWnd =	CreateWindow(GetWindowClassName(), GetWindowName(),
		dwStyle, 0, 0, cx, cy, hParentWnd, NULL, m_hInstance, NULL);

	// Unable to create window, quit.
	if(hScrWnd == NULL) return;

	// Tell windows that a screen saver is running.
	UINT oldval;
	if(GetMode() == smSaver)
		SystemParametersInfo(SPI_SCREENSAVERRUNNING,1,&oldval,0);

	//-----------------------------------------------------
	// STEP 3 - CREATE OPENGL RENDERING CONTEXT
	//-----------------------------------------------------
	HDC hDC = GetDC(hScrWnd);
	// First, we create a pixel format descriptor.
	PIXELFORMATDESCRIPTOR pfd;
    memset(&pfd, 0, sizeof(pfd));
    pfd.nSize        = sizeof(pfd);
    pfd.nVersion     = 1;
    pfd.dwFlags      = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType   = PFD_TYPE_RGBA;
    pfd.cColorBits   = 32;
	// Find a suitable pixel format.
	int pf = ChoosePixelFormat(hDC, &pfd);
    if (pf == 0) 
	{
		MessageBox(NULL, "ChoosePixelFormat() failed:  "
			"Cannot find a suitable pixel format.", "Error", MB_OK); 
		return;
    } 
	// Apply this pixel format to the device context.
    if (SetPixelFormat(hDC, pf, &pfd) == FALSE) 
	{
		MessageBox(NULL, "SetPixelFormat() failed:  "
			"Cannot set format specified.", "Error", MB_OK);
		return;
    } 
    DescribePixelFormat(hDC, pf, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

	// Finally, create the rendering context.
	HGLRC hRC = wglCreateContext(hDC);
	// Then bind it to this thread.
    wglMakeCurrent(hDC, hRC);
	
	//-----------------------------------------------------
	// STEP 4 - INIT THE APPLICATION ITSELF
	//-----------------------------------------------------
	if(!InitGL(cx, cy))
	{
		MessageBox(NULL, "Application Init failed, bailing out...", "Error", MB_OK);
		return;
	}

	//-----------------------------------------------------
	// STEP 5 - SCREEN SAVER MAIN LOOP
	//-----------------------------------------------------
	ShowWindow(hScrWnd, nCmdShow);
	
	MSG msg;

	bool done = false;
	while(!done)									// Loop That Runs While done=FALSE
	{
		if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))	// Is There A Message Waiting?
		{
			if (msg.message==WM_QUIT)				// Have We Received A Quit Message?
				done=TRUE;							// If So done=TRUE
			else									// If Not, Deal With Window Messages
			{
				TranslateMessage(&msg);				// Translate The Message
				DispatchMessage(&msg);				// Dispatch The Message
			}
		}
		else										// If There Are No Messages
		{
			// Update and display application
			Update();
			Display();			
			SwapBuffers(hDC);					// Swap Buffers (Double Buffering)
		}
	}

	//-----------------------------------------------------
	// STEP 6 - RELEASE RESSOURCES
	//-----------------------------------------------------
	ReleaseGL();	// Destroy OpenGL Ressources allocated by the application.
	wglMakeCurrent(NULL, NULL);
    ReleaseDC(hScrWnd, hDC);
    wglDeleteContext(hRC);

	if(GetMode() == smSaver)
		SystemParametersInfo(SPI_SCREENSAVERRUNNING,0,&oldval,0);
	return;
}


// Module Internal Functions ************************************************* //


/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:	SaverWindowProc
// DESCRIPTION:	Yep, the big and ugly window procedure. It handles everything a
//				screen saver should do. 
// PARAMETERS:	...
// RETURN VALUE: ...
// REMARKS:		None
/////////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK SSBaseApp::SaverWindowProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	SSBaseApp *pApp = SSBaseApp::GetGlobalInstance();

	switch (msg)
	{
	case WM_CREATE:		pApp->OnCreate(hwnd);	break;
    case WM_ACTIVATE:
	case WM_ACTIVATEAPP:
	case WM_NCACTIVATE:
		{
			if(	pApp->GetMode() == smSaver && !pApp->m_bIsDialogActive && 
				LOWORD(wParam) == WA_INACTIVE)
			{
				pApp->CloseSaverWindow();
			}
		} break;
	case WM_SETCURSOR:
		{
			if( pApp->GetMode()== smSaver && !pApp->m_bIsDialogActive )
				SetCursor(NULL);
			else
				SetCursor(LoadCursor(NULL,IDC_ARROW));
		} break;
    case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_KEYDOWN:
		{
			if( pApp->GetMode()== smSaver && !pApp->m_bIsDialogActive )
				pApp->CloseSaverWindow();
		} break;
	case WM_MOUSEMOVE:
		{
			if( pApp->GetMode() == smSaver && !pApp->m_bIsDialogActive )
			{
				POINT pt;
				GetCursorPos(&pt);
				int dx = abs(pt.x - pApp->m_InitCursorPos.x);				
				int dy = abs(pt.y - pApp->m_InitCursorPos.y);
				if(dx > (int)pApp->m_dwMouseThreshold || dy > (int)pApp->m_dwMouseThreshold)
				{
					pApp->CloseSaverWindow();
				}
			}
		} 
		break;
	case WM_SYSCOMMAND:
		{
			if( pApp->GetMode() == smSaver )
			{
				if( wParam==SC_SCREENSAVE )
					return false;
				if( wParam==SC_CLOSE )
					return false;
			}
		} 
		break;
	case WM_CLOSE:
		{
			if( pApp->GetMode()== smSaver && pApp->m_bReallyClose && !pApp->m_bIsDialogActive)
			{
				BOOL bCanClose = TRUE;
				if(GetTickCount()-pApp->m_dwInitTime > 1000*pApp->m_dwPasswordDelay )
				{
					pApp->StartDialog();
					bCanClose = SSUtl_VerifyPassword(hwnd);
					pApp->EndDialog();
				}
				if(bCanClose)	DestroyWindow(hwnd);
			}
			if(pApp->GetMode()==SSBaseApp::smSaver)
			{
				// so that DefWindowProc doesn't get called, 
				// because it would just DestroyWindow
				return false; 
			}
		} 
		break;
	case WM_DESTROY:	
		PostQuitMessage(0);
		break;
  }
  return DefWindowProc(hwnd,msg,wParam,lParam);
}
